/*
 * $Header: /home/jerenkrantz/tmp/commons/commons-convert/cvs/home/cvs/jakarta-commons//resources/src/java/org/apache/commons/resources/impl/ResourceBundleResources.java,v 1.15 2004/03/27 21:34:09 dgraham Exp $
 * $Revision: 155432 $
 * $Date: 2005-02-26 08:15:50 -0500 (Sat, 26 Feb 2005) $
 *
 * ====================================================================
 *
 *  Copyright 2003-2004 The Apache Software Foundation
 * 
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

package org.apache.commons.resources.impl;

import java.io.InputStream;
import java.io.Reader;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.TimeZone;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.resources.ResourcesException;
import org.apache.commons.resources.ResourcesKeyException;

/**
 * <p>Concrete implementation of 
 * {@link org.apache.commons.resources.Resources} that wraps a set
 * (one per Locale) of <code>java.util.ResourceBundle</code> instances
 * that share a common base name.  The <code>timeZone</code> argument
 * is ignored in all resource getter method implementations.</p>
 */
public class ResourceBundleResources extends ResourcesBase {

    /**
     * <p>The logging instance for this class.</p>
     */
    private static final Log log =
        LogFactory.getLog(ResourceBundleResources.class);

    // ----------------------------------------------------------- Constructors


    /**
     * <p>Create a new {@link org.apache.commons.resources.Resources} 
     * instance with the specified
     * logical name and bundle base name.</p>
     *
     * @param name Logical name of the new instance
     * @param base Fully qualified base name of the <code>ResourceBundle</code>
     *  instances to be wrapped
     */
    public ResourceBundleResources(String name, String base) {

        super(name);
        this.base = base;

    }


    // ----------------------------------------------------- Instance Variables


    /**
     * <p>The fully qualified base name of the <code>ResourceBundle</code>
     * instances to be wrapped.</p>
     */
    protected String base = null;


    /**
     * <p>Two-level cache of <code>ResourceBundle</code> instances
     * that have previously been acquired by this {@link org.apache.commons.resources.Resources}
     * instance.  The first-level key is the <code>ClassLoader</code>
     * instance used to load the bundles.  The first-level value is a
     * <code>java.util.Map</code> that is then, in turn, keyed by the
     * <code>java.util.Locale</code> to which a particular bundle belongs.</p>
     */
    protected Map bundles = new HashMap();


    // ------------------------------------------------------ Lifecycle Methods


    /**
     * <p>This must be called to initialize the data content of this
     * {@link org.apache.commons.resources.Resources} instance, before any of the <code>getXxx()</code>
     * methods are called.</p>
     *
     * @exception ResourcesException if an error occurs during initialization
     */
    public void init() throws ResourcesException {

        if (log.isDebugEnabled()) {
            log.debug("Initializing for base name '" + base + "'");
        }

    }


    /**
     * <p>This method must be called when the manager of this resource
     * decides that it's no longer needed.  After this method is called,
     * no further calls to any of the <code>getXxx()</code> methods are
     * allowed.</p>
     *
     * @exception ResourcesException if an error occurs during finalization
     */
    public void destroy() throws ResourcesException {

        if (log.isDebugEnabled()) {
            log.debug("Finalizing for base name '" + base + "'");
        }
        
        synchronized (bundles) {
            Iterator loaders = bundles.keySet().iterator();
            while (loaders.hasNext()) {
                ClassLoader loader = (ClassLoader) loaders.next();
                ((Map) bundles.get(loader)).clear();
            }
            bundles.clear();
        }

    }


    // ------------------------------------------------------------- Properties


    /**
     * <p>Return the fully qualified base name of the
     * <code>ResourceBundle</code> instances we are wrapping.</p>
     */
    public String getBase() {

        return (this.base);

    }


    public Iterator getKeys() {

        synchronized (bundles) {

            ClassLoader loader = getClassLoader();
            Map loaderMap = (Map) bundles.get(loader);
            if (loaderMap == null) {
                return (Collections.EMPTY_SET.iterator());
            }
            Set results = new HashSet();
            Iterator locales = loaderMap.keySet().iterator();
            while (locales.hasNext()) {
                Locale locale = (Locale) locales.next();
                ResourceBundle bundle = (ResourceBundle)
                    loaderMap.get(locale);
                Enumeration keys = bundle.getKeys();
                while (keys.hasMoreElements()) {
                    results.add(keys.nextElement());
                }
            }
            return (results.iterator());

        }

    }


    // ---------------------------------------------- Content Retrieval Methods


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * byte array, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public byte[] getBytes(String key, Locale locale, TimeZone timeZone)
        throws ResourcesException {

        throw new UnsupportedOperationException(); // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as an
     * InputStream, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public InputStream getInputStream(String key, Locale locale,
                                      TimeZone timeZone) {

        throw new UnsupportedOperationException();  // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as an
     * Object, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public Object getObject(String key, Locale locale, TimeZone timeZone) {

        try {
            ResourceBundle bundle = getBundle(locale, timeZone);
            return (bundle.getObject(key));
            
        } catch (MissingResourceException e) {
            if (isReturnNull()) {
                return (null);
            } else {
                throw new ResourcesKeyException(key);
            }
        }

    }


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * Reader, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public Reader getReader(String key, Locale locale, TimeZone timeZone) {

        throw new UnsupportedOperationException(); // FIXME

    }


    /**
     * <p>Return the content for the specified <code>key</code> as a
     * String, localized based on the specified <code>locale</code>
     * and/or <code>timeZone</code>.</p>
     *
     * @param key Identifier for the requested content
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception ResourcesException if an error occurs retrieving or
     *  returning the requested content
     * @exception ResourcesKeyException if the no value for the specified
     *  key was found, and <code>isReturnNull()</code> returns
     *  <code>false</code>
     */
    public String getString(String key, Locale locale, TimeZone timeZone) {

        if (log.isTraceEnabled()) {
            log.trace("Retrieving message for key '" + key + "' for locale '"
                      + locale + "'");
        }
        
        try {
            ResourceBundle bundle = getBundle(locale, timeZone);
            String message = bundle.getString(key);
            if (log.isTraceEnabled()) {
                log.trace("Retrieved message is '" + message + "'");
            }
            return (message);
            
        } catch (ClassCastException e) {
            throw new ResourcesException(e);
            
        } catch (MissingResourceException e) {
            if (log.isTraceEnabled()) {
                log.trace("No message found");
            }
            if (isReturnNull()) {
                return (null);
            } else {
                throw new ResourcesKeyException(key);
            }
        }

    }



    // ------------------------------------------------------ Protected Methods


    /**
     * <p>Return the appropriate <code>ResourceBundle</code> instance
     * that corresponds to the specified <code>locale</code> and
     * <code>timeZone</code> parameters.  The first time a particular
     * bundle is requested, cache it so that subsequent requests will
     * operate more quickly.</p>
     *
     * <p>The default implementation of this method unconditionally ignores the
     * <code>timeZone</code> argument.</p>
     *
     * @param locale Locale with which to localize retrieval,
     *  or <code>null</code> for the default Locale
     * @param timeZone TimeZone with which to localize retrieval,
     *  or <code>null</code> for the default TimeZone
     *
     * @exception MissingResourceException if the requested Resourcebundle
     *  cannot be acquired
     */
    protected ResourceBundle getBundle(Locale locale, TimeZone timeZone)
        throws MissingResourceException {

        if (locale == null) {
            locale = Locale.getDefault();
        }

        synchronized (bundles) {

            // Locate or create the Map for the appropriate ClassLoader
            ClassLoader loader = getClassLoader();
            Map loaderMap = (Map) bundles.get(loader);
            if (loaderMap == null) {
                loaderMap = new HashMap();
                bundles.put(loader, loaderMap);
            }

            // Locate or create the requested ResourceBundle instance
            ResourceBundle bundle =
                (ResourceBundle) loaderMap.get(locale);
            if (bundle == null) {
                bundle = ResourceBundle.getBundle(base, locale, loader);
                loaderMap.put(locale, bundle);
            }
            return (bundle);

        }

    }


    /**
     * <p>Return the <code>ClassLoader</code> for which we are mapping
     * <code>ResourceBundle</code> instances.</p>
     */
    protected ClassLoader getClassLoader() {

		ClassLoader loader = Thread.currentThread().getContextClassLoader();
        
        if (loader == null) {
            loader = this.getClass().getClassLoader();
        }
        
        return (loader);

    }


}
